use std::str::FromStr;

use thiserror::Error;

use super::{seed_range::SeedRange, UnsignedNumber};

#[derive(Debug, PartialEq, Eq, Clone)]
pub struct ItemMapping {
    source_range: SeedRange,
    target: UnsignedNumber,
    range: UnsignedNumber,
}

impl ItemMapping {

    pub fn map_range(&self, to_map: SeedRange) -> MappedRange {
        let start = self.source_range.start();
        let end = self.source_range.end();
        // xxxxxx => to_map
        // yyyyyy => self

        // -------xxxxxx
        // yyyyyy-------
        if to_map.start() > end || 
        // xxxxxxx---------
        // ----------yyyyyy
        to_map.end() < start {
            MappedRange::Outside
        } else if to_map.start() < start && to_map.end() <= end {
            // xxxxxxx----
            // ----yyyyyyy
            let outside_start = to_map.start();
            let outside_end =  start.saturating_sub(1);
            let outside = SeedRange::new(outside_start, outside_end);
            
            let mapped_start = self.target;
            let mapped_end = self.offset_mapping(to_map.end());
            let mapped = SeedRange::new(mapped_start, mapped_end);
            MappedRange::PartiallyWithin { outside, mapped }
        } else if to_map.start() >= start && to_map.end() > end {
            // yyyyyy----
            // ----xxxxxx
            let outside_start = end + 1;
            let outside_end = to_map.end();
            let outside = SeedRange::new(outside_start, outside_end);
            
            let mapped_start = self.offset_mapping(to_map.start());
            let mapped_end = self.offset_mapping(outside_start - 1);
            let mapped = SeedRange::new(mapped_start, mapped_end);
            MappedRange::PartiallyWithin { outside, mapped }
            
        } else if to_map.start() < start && to_map.end() > end {
            // -----yyy----
            // ----xxxxxx--
            self.map_from_left_and_right(to_map)
        } else {
            // ----yyyyyy----
            // -----xxxx-----
            let mapped_start = self.offset_mapping(to_map.start());
            let mapped_end = self.offset_mapping(to_map.end());
            let mapped = SeedRange::new(mapped_start, mapped_end);
            MappedRange::Within(mapped)
        }
    }

    pub fn map_point(&self, to_map: UnsignedNumber) -> Option<UnsignedNumber> {
        let source_range = self.source_range;
        let start = source_range.start();
        let end = source_range.end(); if start <= to_map && end >= to_map {
            let mapped = self.offset_mapping(to_map);
            Some(mapped)
        } else {
            None
        }
    }

    fn map_from_left_and_right(&self, to_map: SeedRange) -> MappedRange {
        let (start, end) = self.source_start_end();
        let left = {
            let left_start = to_map.start();
            let left_end = start - 1;
            SeedRange::new(left_start, left_end)
        };
        let right = {
            let right_start = end + 1;
            let right_end = to_map.end();
            SeedRange::new(right_start, right_end)
        };
        let mapped = {
            let mapped_start = self.target;
            let mapped_end = self.offset_mapping(end);
            SeedRange::new(mapped_start, mapped_end)
        };
        MappedRange::FromBothSidePartiallyWithin { outside: (left, right), mapped }
    }
    
    fn source_start_end(&self) -> (UnsignedNumber, UnsignedNumber) {
        let range = self.source_range;
        (range.start(), range.end())
    }
    
    fn offset_mapping(&self, to_map: UnsignedNumber) -> UnsignedNumber {
        assert!(to_map >= self.source_range.start());
        let difference = to_map - self.source_range.start();
        self.target + difference
    }
}

#[derive(Debug, PartialEq, Eq)]
pub enum MappedRange {
    Outside,
    Within(SeedRange),
    PartiallyWithin {
        outside: SeedRange,
        mapped: SeedRange,
    },
    FromBothSidePartiallyWithin {
        outside: (SeedRange, SeedRange),
        mapped: SeedRange,
    },
}

#[derive(Debug, Error, PartialEq, Eq)]
#[error("Item mapping must be exactly 3 numbers")]
pub struct InvalidStrItemMapping;

impl FromStr for ItemMapping {
    type Err = InvalidStrItemMapping;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        let mut splitted = s.split_whitespace().take(3).map(|to_parse| {
            to_parse
                .parse::<UnsignedNumber>()
                .map_err(|_| InvalidStrItemMapping)
        });
        match (splitted.next(), splitted.next(), splitted.next()) {
            (Some(Ok(target)), Some(Ok(source)), Some(Ok(range))) => {
                let inclusive_range = range.saturating_sub(1);
                let end = source + inclusive_range;
                let source_range = SeedRange::new(source, end);
                let parsed_self = Self {
                    source_range,
                    target,
                    range,
                };
                Ok(parsed_self)
            }
            (_, _, _) => Err(InvalidStrItemMapping),
        }
    }
}

#[cfg(test)]
mod testing {
    use crate::solutions::day5::{
        item_mapping::{InvalidStrItemMapping, ItemMapping, MappedRange},
        seed_range::SeedRange,
        UnsignedNumber,
    };

    #[test]
    fn from_str_item_mapping() {
        fn assert_case(input: &str, expected: Result<ItemMapping, InvalidStrItemMapping>) {
            let actual = input.parse();
            assert_eq!(expected, actual, "Input: {}", input);
        }
        assert_case(
            "20 10 22",
            Ok(ItemMapping {
                source_range: SeedRange::new(10, 31),
                target: 20,
                range: 22,
            }),
        );
        assert_case(
            "50 98 2",
            Ok(ItemMapping {
                source_range: SeedRange::new(98, 99),
                target: 50,
                range: 2,
            }),
        );
        assert_case(
            "52 50 48",
            Ok(ItemMapping {
                source_range: SeedRange::new(50, 97),
                target: 52,
                range: 48,
            }),
        );
        assert_case("a 10 22", Err(InvalidStrItemMapping));
        assert_case("10 10", Err(InvalidStrItemMapping));
    }
    #[test]
    fn should_map_point() {
        fn assert_case(point: UnsignedNumber, input: &str, expected: Option<UnsignedNumber>) {
            let mapping: ItemMapping = input.parse().unwrap();
            let actual = mapping.map_point(point);

            assert_eq!(
                expected, actual,
                "Point: {:?}\nMaping: {:?}",
                point, mapping
            );
        }

        assert_case(2, "4 2 2", Some(4));
        assert_case(0, "4 2 2", None);
        assert_case(4, "4 2 2", None);
        assert_case(79, "52 50 48", Some(81));
    }

    #[test]
    fn should_map_ranges() {
        fn assert_case(given: SeedRange, mapping: ItemMapping, expected: MappedRange) {
            let actual = mapping.map_range(given);
            assert_eq!(
                expected, actual,
                "Given: {:?}\nMapping: {:?}",
                given, mapping
            );
        }
        let example_mapping = ItemMapping {
            source_range: SeedRange::new(98, 99),
            target: 50,
            range: 2,
        };

        assert_case(
            SeedRange::new(95, 97),
            example_mapping.clone(),
            MappedRange::Outside,
        );
        assert_case(
            SeedRange::new(100, 120),
            example_mapping.clone(),
            MappedRange::Outside,
        );
        assert_case(
            SeedRange::new(98, 99),
            example_mapping.clone(),
            MappedRange::Within(SeedRange::new(50, 51)),
        );
        let greater_example_mapping = ItemMapping {
            source_range: SeedRange::new(50, 97),
            target: 52,
            range: 48,
        };
        assert_case(
            SeedRange::new(40, 80),
            greater_example_mapping.clone(),
            MappedRange::PartiallyWithin {
                outside: SeedRange::new(40, 49),
                mapped: SeedRange::new(52, 82),
            },
        );
        assert_case(
            SeedRange::new(80, 100),
            greater_example_mapping.clone(),
            MappedRange::PartiallyWithin {
                outside: SeedRange::new(98, 100),
                mapped: SeedRange::new(82, 99),
            },
        );
        assert_case(
            SeedRange::new(30, 100),
            greater_example_mapping.clone(),
            MappedRange::FromBothSidePartiallyWithin {
                outside: (SeedRange::new(30, 49), SeedRange::new(98, 100)),
                mapped: SeedRange::new(52, 99),
            },
        );
        assert_case(
            SeedRange::new(96, 102), 
            example_mapping.clone(), 
            MappedRange::FromBothSidePartiallyWithin {
                outside: (SeedRange::new(96, 97), SeedRange::new(100, 102)),
                mapped: SeedRange::new(50, 51),
            },
        );
    }
}
